/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**          AnyX.c (previously named Unix.c)               **/
/**                                                         **/
/** This file contains improved Unix/X-dependent            **/
/** subroutines and drivers. Functions in Refresh.c are     **/
/** used to build ximages, functions in Colors.c handle     **/
/** allocating and choosing colors.                         **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin  1994,1995                **/
/**               Arnold Metselaar 1998-2004                **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify us, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/


/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "MSX.h"
#include "AnyX.h"
#include "Keyboard.h"
#include "Sound.h"

/** MIT Shared Memory Extension for X ************************/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
static XShmSegmentInfo SHMInfo;
int UseSHM=1;
/** The following is used for MITSHM-auto-detection: **/
static int ShmOpcode;       /* major opcode for MIT-SHM */
static int (*OldHandler)();
#endif

/** Sound-related declarations *******************************/
#ifdef SOUND
int SndSwitch = (1<<MAXCHANNELS)-1;
int SndVolume = 255/MAXCHANNELS;
int UseSound=44100;
#endif

#if 0 /* needs fixing */
#define SND_INIT 1 /* is InitSound() called succesfully */
#define SND_REQ  2 /* is Sound requested */
#define SND_ON   4 /* is the sound on */
#endif

/****************************************************************/
/*** These dummy functions are called when writes to sound    ***/
/*** registers occur. Replace them with the actual functions  ***/
/*** generating sound.                                        ***/
/****************************************************************/
#ifndef SOUND
/* int UseSound; */
void PSGOut(register byte R,register byte V)  { }
void SCCOut(register byte R,register byte V)  { }
#endif
void OPLLOut(register byte R,register byte V) { }

typedef struct {
    int b_next;  /* next line where settings change */
    int b_width; /* screen-width for this portion of the screen */
    int b_dHor;  /* horizontal adjust for this portion of the screen */
    XID b_C0,b_C1; /* border colors, unswapped */
} Border; 

extern void DOption();
char *Title="fMSX Unix/X 2.8/3.5";

/** Variables to pass options to the X-drivers ***************/
int SyncFreq=0;
int SaveCPU=1;
int UseZoom;
/* handled internally */
static char *Dispname=NULL;
static int do_mono=0;
static int do_install, ncol; /* own colormap, #entries in std map */
static int do_sync;

/** function used during the processing of XtraOptions[] */
static void SetIFreq(int);
static void SetDepth(int);
static void SetXScreen(int);
static void SetVisual(char *);
static void SetVisualId(long);



/* DOption() may want to change this */
int CurDisk[2];
char *Disks[2][MAXDISKS+1];

/* For internal use */
static volatile int CFlag=0; /* Command mode requested? */
static volatile byte TimerReady;
int FastForward;
static int IntSync=-2;
static byte XKeyMap[16];
static byte JoyState;

/** X-related definitions ************************************/
#define KeyMask KeyPressMask|KeyReleaseMask|KeymapStateMask
#define OtherMask FocusChangeMask|ExposureMask|StructureNotifyMask
#define MyEventMask KeyMask|OtherMask

/** Various X-related variables ******************************/
Display *Dsp;
int bitpix;
Window Wnd;
static Atom DELETE_WINDOW_atom, WM_PROTOCOLS_atom;
static Pixmap MyStipple;
static Colormap CMap;
static int CMapIsMine=0;
static GC gcrand,gcbw;
XVisualInfo VisInfo;
static long VisInfo_mask;
static XImage *ximage;
byte *XBuf;
static Border brdrstat[212];
static XGCValues gcvalues;

int isgcrand, isgcbw, isMyStipple, Mapped; /* flags */

/* size of window */
Bool wide,high;
static int Width,Height; /* current window size */
static int BrdrLine;     /* last line where border changed */
static int Lines, LastLine, dVer; /* Vertical adjust */

#define xWIDTH(x)  ((x)?552:276)
#define xHEIGHT(x) ((x)?456:228) /* nice values for window size */

/** Screen Mode Handlers [number of screens + 1] *************/
#if 0
void (*RefreshLine[MAXSCREEN+2])(byte Y) =
{
   RefreshLine0,   /*SCR0:TEXT 40x24 */
   RefreshLine1,   /*SCR1:TEXT 32x24 */
   RefreshLine2,   /*SCR2:BLK 256x192*/
   RefreshLine3,   /*SCR3: 64x 48x 16*/
   RefreshLine4,   /*SCR4:BLK 256x192*/
   RefreshLine5,   /*SCR5:256x192x 16*/
   RefreshLine6,   /*SCR6:512x192x  4*/
   RefreshLine7,   /*SCR7:512x192x 16*/
   RefreshLine8,   /*SCR8:256x192x256*/
   RefreshLineTx80 /*SCR0:TEXT 80x24 */
}; 
#endif

static void 
NewBlock(int start)
{
  int i;

  for ( i=BrdrLine ; i<start ; ++i )
    brdrstat[i].b_next=start;

  BrdrLine=start;
}

void 
BorderChange(int y, ColTyp c0, ColTyp c1, int width, int offset)
{
  int i,dHor;
  XID C0,C1;

  LastLine=y;
  if (y==0) {
    dVer=(ScanLines212 ? 212/2 : 192/2) + 8-(((VDP[18]>>4)&0x0F)^0x08);
    Used8=0;
  }
  dHor=8-((VDP[18]&0x0F)^0x08);
  if (width==0)
    brdrstat[y].b_width=brdrstat[y].b_dHor=0;
  else
    {
      if (VDP[25]&0x02 /* MSK */) {
	brdrstat[y].b_width=width-8;
	brdrstat[y].b_dHor = ((8+dHor+offset)*256 /* offset */
			      +8-(VDP[27]&0x07)); /* skip */
      } else {
	brdrstat[y].b_width=width-(VDP[27]&0x07);
	brdrstat[y].b_dHor = ((dHor+offset+(VDP[27]&0x07))*256 /* offset */
			      +0);/* skip */
      }
    }
  if (do_mono) {
    brdrstat[y].b_C0 = C0 = (c0.ct_byte?white:black);
    brdrstat[y].b_C1 = C1 = (c1.ct_byte?white:black);
  } else {
    brdrstat[y].b_C0 = C0 = UnSwap(c0);
    brdrstat[y].b_C1 = C1 = UnSwap(c1);
  }

  if (y==(ScanLines212? 211:191)) {
    NewBlock(y+1);
    for ( i=y+1 ; i<212 ; ++i)
      brdrstat[i].b_width=0;
    NewBlock(212);
  }
  else {
    if ( (brdrstat[y+1].b_width!=brdrstat[y].b_width) || 
	 (brdrstat[y+1].b_dHor!=brdrstat[y].b_dHor) ||
	 (brdrstat[y+1].b_C0!=C0)||
	 (brdrstat[y+1].b_C1!=C1) )
      brdrstat[y].b_next=y+1;
    else
      brdrstat[y].b_next=brdrstat[y+1].b_next;
  }
  if ((brdrstat[BrdrLine].b_width!=brdrstat[y].b_width) || 
      (brdrstat[BrdrLine].b_dHor!=brdrstat[y].b_dHor) ||
      (brdrstat[BrdrLine].b_C0!=C0)||
      (brdrstat[BrdrLine].b_C1!=C1))
    NewBlock(y);
  else
    brdrstat[BrdrLine].b_next=brdrstat[y].b_next;
}

/** ChangeGC: function setting gcrand according to brdrstat *****/
static void 
ChangeGC(register int y) 
{
  XID C0,C1;
  if (Width & 2)
    { C0=brdrstat[y].b_C1; C1=brdrstat[y].b_C0; }
  else
    { C0=brdrstat[y].b_C0; C1=brdrstat[y].b_C1; }
  
  if ((gcvalues.foreground!=C1)||(gcvalues.background!=C0)) {
    gcvalues.foreground=C1; 
    gcvalues.background=C0;
    XChangeGC(Dsp,gcrand,GCForeground|GCBackground,&gcvalues);
  }
}      

/** PUTIMAGE: functon copying image buffer into a window ********/
void 
PutImage(void)
{ 
  int y,ysc,ywin,xsc;
  int bwid,bhig;

  ysc=(high?2:1);
  xsc=(wide?2:1);
  ywin=Height/2-ysc*dVer;

  ChangeGC(0);
  if (ywin>0)
    XFillRectangle(Dsp,Wnd,gcrand, 0,0, Width,ywin);

  for (y=0 ; y<Lines ; ywin+=bhig, y=brdrstat[y].b_next) 
    {
      if (y)  ChangeGC(y);
      bhig=ysc*(brdrstat[y].b_next-y);
      if (brdrstat[y].b_width)
	  {
	    int offset, skip, dwidth;

	    offset=brdrstat[y].b_dHor/256;
	    skip=brdrstat[y].b_dHor & 0xFF;

	    bwid=Width/2+xsc*(offset-128);
	    if (bwid>0)
	      XFillRectangle(Dsp,Wnd,gcrand, 0,ywin, bwid,bhig);
	    dwidth=xsc*brdrstat[y].b_width;
#ifdef MITSHM
	    if (UseSHM)
		XShmPutImage(Dsp,Wnd,gcbw,ximage, xsc*skip,ysc*y,
			     bwid,ywin, dwidth,bhig, False);
	    else                                                       
#endif            
	      XPutImage(Dsp,Wnd,gcbw,ximage, xsc*skip,ysc*y,
			bwid,ywin, dwidth,bhig);
	    bwid=Width-bwid-dwidth;
	    if (bwid>0)
	      XFillRectangle(Dsp,Wnd,gcrand, Width-bwid,ywin, bwid,bhig);
	  }
      else
	  XFillRectangle(Dsp,Wnd,gcrand, 0,ywin, Width,bhig);
    }
  if (Height>ywin)
    XFillRectangle(Dsp,Wnd,gcrand, 0,ywin, Width,Height-ywin);
  XFlush(Dsp);
}

void 
RefreshScreen(void)
{
  int i,msk;

  /* Wait if needed */
  if(!IntSync && !FastForward) 
    while(!TimerReady&&!ExitNow)
      if(CFlag) { DOption(); CFlag=0; }
  if (!Mapped) return;
  Lines=LastLine+1;
  if (brdrstat[BrdrLine+1].b_next < brdrstat[BrdrLine].b_next)
    for (i=BrdrLine ; i < Lines ; ++i)
      brdrstat[i].b_next=Lines;
  BrdrLine=0;
  eo_flag ^= 1;
  PutImage();
  msk=0xffff & ~(PalUsed | PalOut);
  if (msk)
    {
      Colors.Leave(msk);
      PalOut|=msk;
    }
  if (! Used8)
    Colors.LeaveScr8();
  Colors.Refresh();
  /* Reset timer flag */
  TimerReady=0;
}

void 
OnBreak(int ARG) 
{ 
  ExitNow=1; 
}
void 
OnInt(int ARG)
{
  CFlag=1; 
  signal(ARG,OnInt);
}

void 
OnTimer(int ARG) 
{
  TimerReady=1; 
  signal(ARG,OnTimer);
}

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void 
TrashMachine(void)
{
  if(Verbose & 1) printf("Shutting down...\n");

  TrashColor();
  if(Dsp&&Wnd)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(Dsp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
#endif MITSHM
    if(ximage) XDestroyImage(ximage);
    if(isgcrand) XFreeGC(Dsp,gcrand);
    if(isgcbw) XFreeGC(Dsp,gcbw);
    if(isMyStipple) XFreePixmap(Dsp,MyStipple);
    XAutoRepeatOn(Dsp);
  }
  if(CMapIsMine) XFreeColormap(Dsp, CMap);
  if(Dsp) XCloseDisplay(Dsp);


#ifdef SOUND
  StopSound(); TrashSound();
#endif SOUND
}

static struct vc {
    int clasnum;
    char *name;
} visualClasses[6]=
  {
      { StaticGray, "StaticGray"},
      { GrayScale, "GrayScale"},
      { StaticColor, "StaticColor"},
      { PseudoColor, "PseudoColor"},
      { TrueColor, "TrueColor"},
      { DirectColor, "DirectColor"}
  };

void 
SetDepth(int x)
{
    VisInfo.depth=x;
    VisInfo_mask |= VisualDepthMask;
}

void 
SetXScreen(int x)
{
    VisInfo.screen=x;
    VisInfo_mask |= VisualScreenMask;
}

void 
SetVisual(char *s)
{
    int i;
    char *vc, *sp;

    for (i=0; i<6; i++) {
        for (  vc=visualClasses[i].name, sp=s ; (*vc!=0) ; vc++, sp++ )
	    if (tolower(*vc)!=tolower(*sp))
		goto next;
	VisInfo.class=visualClasses[i].clasnum;
	VisInfo_mask |= VisualClassMask;
	return;
      next: ;
    }
    printf("%s is not a known visual class.\n",s);
}

void 
SetVisualId(long x)
{
    VisInfo.visualid=(VisualID) x;
    VisInfo_mask |= VisualIDMask;
}

int 
ChooseVisual(int *do_install, int *do_mono)
{
    int n_items,i;
    XVisualInfo *VisInfo_l;

    if (Verbose&1) printf("OK\n  Choosing Visual...");

    if (VisInfo_mask!=VisualNoMask) {
	VisInfo_l=XGetVisualInfo(Dsp, VisInfo_mask, &VisInfo, &n_items);
	if (n_items!=0) {
	    if ( (Verbose&1) && (n_items>1)) 
		printf("%d matching visuals...",n_items);
	    VisInfo=VisInfo_l[0]; /* this may not be the best match */
	    XFree(VisInfo_l);
	    goto found;
	} else 
	    if (Verbose&1)
		printf("no matching visual...");
    }

    if (Verbose&1)
	printf("using default...");

    VisInfo.screen= DefaultScreen(Dsp);
    VisInfo.visualid=DefaultVisual(Dsp, VisInfo.screen)->visualid;
    VisInfo_l=XGetVisualInfo(Dsp, VisualIDMask|VisualScreenMask, 
			     &VisInfo, &n_items);
    if (n_items!=0) {
	VisInfo=VisInfo_l[0]; /* this is the only match */
	XFree(VisInfo_l);
	goto found;
    }

    if (Verbose&1)
	printf("FAILED\n");
    return 0;

    found:

    if (Verbose&1) {
	for (i=6 ; i ; )
	    if (visualClasses[--i].clasnum==VisInfo.class)
		break;	
	printf("\n    visualid=0x%lx: %s, ", VisInfo.visualid, 
	       (i>=0) ? visualClasses[i].name : "unknown");
    }
    if (VisInfo.visualid != DefaultVisual(Dsp,VisInfo.screen)->visualid)
	*do_install=1;
    if (*do_install)
	*do_mono=0;

    return 1;
}

/* Function to choose a PixMapFormat and set some variables associated
   with them.  Note that the depth comes from a visual and is
   therefore supported by the screen */
int 
ChooseFormat()
{
  int i;
  int nForm; /* # pixmapformats, depth */
  XPixmapFormatValues *flist, choice; /* list of format describers
					 and choice */

  flist=XListPixmapFormats(Dsp,&nForm);
  if (!flist) return (0); 

  choice.bits_per_pixel=choice.depth=0;
  for (i=0;i<nForm;i++)
    {
      if (flist[i].depth==VisInfo.depth)
	  choice=flist[i];
    }
	     
  XFree((void *)flist);
  if (choice.depth==0) {
      if (Verbose &1) printf("FAILED\n"); 
      return (0);
  }

  bitpix=choice.bits_per_pixel; 
  return 1;
}

#ifdef MITSHM
/* Error handler, used only to trap BadAccess errors on XShmAttach() */
int 
Handler(Display * d, XErrorEvent * Ev)
{ 
  if ( (Ev->error_code==10) /* BadAccess */ && 
       (Ev->request_code==(unsigned char) ShmOpcode) &&
       (Ev->minor_code==1)  /* X_ShmAttach */ )
    UseSHM=0; /* X-terminal apparently remote */
  else return OldHandler(d,Ev);
  return 0;
}

/* Function that tries to XShmAttach an XImage, returns 0 in case of
   faillure */
int 
TrySHM(int depth, int screen_num)
{
  long size;

  if(Verbose & 1)
    printf("  Using shared memory:\n    Creating image...");
  ximage=XShmCreateImage(Dsp,VisInfo.visual,
			 do_mono ? 1 : depth,
                         do_mono ? XYBitmap : ZPixmap,
			 NULL, &SHMInfo, WIDTH, HEIGHT);
  if(!ximage) return (0); 

  size=(long)ximage->bytes_per_line*ximage->height;
  if(Verbose & 1) printf("OK\n    Getting SHM info for %ld %s...",
			 (size&0x03ff) ? size : size>>10,
			 (size&0x03ff) ? "bytes" : "kB" );

  SHMInfo.shmid=shmget(IPC_PRIVATE, size, IPC_CREAT|0777);
  if(SHMInfo.shmid<0) return (0);

  if(Verbose & 1) printf("OK\n    Allocating SHM...");
  XBuf=(byte *)(ximage->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
  if(!XBuf) return (0);

  SHMInfo.readOnly=False;
  if(Verbose & 1) printf("OK\n    Attaching SHM...");
  OldHandler=XSetErrorHandler(Handler);
  if (!XShmAttach(Dsp,&SHMInfo)) return(0);
  XSync(Dsp,False);
  XSetErrorHandler(OldHandler);
  return(UseSHM);
}
#endif

/* Inform window manager about the window size, attributes and other
   hints */
void 
SetWMHints(void) 
{
    XTextProperty win_name,icon_name;
    char * value[1];
    XSizeHints *size;
    XWMHints *hint;
    XClassHint *klas;
    char *argv[]= {NULL};
    int argc=0;
    
    size=XAllocSizeHints();
    hint=XAllocWMHints();
    klas=XAllocClassHint();
    
    size->flags=PMinSize;
    size->min_width=xWIDTH(0);
    size->min_height=xHEIGHT(0);
    
    hint->input=True;
    hint->flags=InputHint;
    
    value[0]=Title;
    XStringListToTextProperty(value, 1, &win_name);
    value[0]="fMSX";
    XStringListToTextProperty(value, 1, &icon_name);
    
    klas->res_name="fmsx";
    klas->res_class="Fmsx";
    
    XSetWMProperties(Dsp, Wnd, &win_name, &icon_name, argv, argc,
		     size, hint, klas);
    XFree(size); 
    XFree(hint);
    XFree(klas);

    DELETE_WINDOW_atom=XInternAtom(Dsp, "WM_DELETE_WINDOW", False);
    WM_PROTOCOLS_atom =XInternAtom(Dsp, "WM_PROTOCOLS", False);

    XChangeProperty(Dsp, Wnd, WM_PROTOCOLS_atom , XA_ATOM, 32,
		    PropModeReplace, (unsigned char*) &DELETE_WINDOW_atom,1);

}

/** Make a GC and a Pixmap that can be used by border and PutImage to
    make a nice border **/
GC 
MakeGCrand(void)
{
  int swid,shig,i;
  GC gc;

  XQueryBestStipple(Dsp,Wnd,2,1,&swid,&shig);
  MyStipple=XCreatePixmap(Dsp, Wnd, swid, shig, 1);
  isMyStipple=1;

  if (Verbose) printf("  Dimensions of MyStipple: %d by %d\n",swid,shig);
  gcvalues.function=GXcopy;
  gcvalues.plane_mask=AllPlanes;
  gcvalues.line_width=1;
  gcvalues.line_style=LineSolid;
  gcvalues.cap_style=CapRound;
  gcvalues.fill_style=FillSolid;

  gcvalues.foreground=0;
  gc=XCreateGC(Dsp,MyStipple,GCFunction|GCPlaneMask
	       |GCLineWidth|GCLineStyle|GCCapStyle|GCFillStyle
	       |GCForeground,&gcvalues);
  for (i=0; i<swid; i+=2) XDrawLine(Dsp, MyStipple, gc, i, 0, i, shig-1);

  gcvalues.foreground=1;
  XChangeGC(Dsp, gc, GCForeground , &gcvalues);
  for (i=1; i<swid; i+=2) XDrawLine(Dsp, MyStipple, gc, i, 0, i, shig-1);
  XFreeGC(Dsp,gc);

  gcvalues.function=GXcopy;
  gcvalues.plane_mask=AllPlanes;
  gcvalues.foreground=gcvalues.background=black;
  gcvalues.fill_style=FillOpaqueStippled;
  gcvalues.stipple=MyStipple;
  gcvalues.ts_x_origin=0;
  return XCreateGC(Dsp,Wnd,GCFunction|GCPlaneMask|GCForeground|GCBackground
	       |GCFillStyle|GCStipple|GCTileStipXOrigin, &gcvalues);
}

GC 
MakeGCbw(void)
{
  gcvalues.function=GXcopy;
  gcvalues.plane_mask=AllPlanes;
  gcvalues.foreground=white;
  gcvalues.background=black;
  return XCreateGC(Dsp,Wnd,
		   GCFunction|GCPlaneMask|GCForeground|GCBackground,&gcvalues);
}

void 
SetIFreq(int Hz)
{
  static struct itimerval TimerValue;
  
  signal(SIGALRM,OnTimer);
  if(Verbose & 1) printf("Initializing timer for %dHz...",Hz);
  TimerValue.it_interval.tv_sec =
    TimerValue.it_value.tv_sec    = 0L;
  TimerValue.it_interval.tv_usec=
    TimerValue.it_value.tv_usec   = Hz?1000000L/Hz:0;
  IntSync=setitimer(ITIMER_REAL,&TimerValue,NULL);
  if(Verbose & 1) printf(IntSync?"FAILED\n":"OK\n");
} 

/** MachineOption ********************************************/
/** chack whether an option is specific to these drivers    **/
/*************************************************************/
static char *XtraOptions[]=
{
  "allspr", "normspr", "limspr", "ncol", "install", "noinstall",
  "depth", "visual", "visualid", "display", "screen", "synchronise", 
  "bw", "display", 0
};
int
MachineOption(int argc, char *argv[], int *p)
{
  int J;
  
  if (KeyboardOption(argc,argv,p))
    return 1;
  for(J=0 ; XtraOptions[J] ; J++)
    if(!strcmp(argv[(*p)]+1,XtraOptions[J])) break;
  switch (J) 
    {
    case 0: NSprMax_M=NSprMax_C=32; break;
    case 1: NSprMax_M=4; NSprMax_C=8; break;
    case 2: (*p)++;
            if((*p)<argc) 
	      NSprMax_M=NSprMax_C=atoi(argv[(*p)]); 
	    else printf("%s: No number of sprites per line\n",argv[0]);
	    break;
    case 3: (*p)++;
            if((*p)<argc) { 
	      ncol=atoi(argv[(*p)]);
	      if ((ncol<2)||(ncol>256)) ncol=64;
	    }
	    else printf("%s: No number of colors supplied\n",argv[0]);
	    break;
    case 4: do_install=1; break;
    case 5: do_install=0; break;
    case 6: (*p)++;
            if ((*p)<argc) 
	      SetDepth(atoi(argv[(*p)]));
	    else
	      printf("%s: No depth supplied\n",argv[0]);
	    break;
    case 7: (*p)++;
            if ((*p)<argc) 
	      SetVisual(argv[(*p)]);
	    else
	      printf("%s: No visual-class supplied\n",argv[0]);
	    break;
    case 8: (*p)++;
            if ((*p)<argc)
	      SetVisualId(strtol(argv[(*p)],NULL,0));
	    else
	      printf("%s: No visual-id supplied\n",argv[0]);
	    break;
    case 9: (*p)++;
            if ((*p)<argc)
	      Dispname=argv[(*p)];
	    else
	      printf("%s: No visual-id supplied\n",argv[0]);
	    break;
    case 10: (*p)++;
             if ((*p)<argc) 
	       SetXScreen(atoi(argv[(*p)]));
	     else
	       printf("%s: No screen supplied\n",argv[0]);
	     break;
    case 11: do_sync=1; break;
    case 12: do_mono=1; break;
    case 13: (*p)++;
             if ((*p)<argc) 
	       Dispname=argv[(*p)];
	     else
	       printf("%s: No display-name supplied\n",argv[0]);
	     break;
    }
  return (XtraOptions[J]!=NULL);
}
/** MachineCommand *******************************************/
/** chack whether a command is specific to these drivers    **/
/*************************************************************/
static char *XtraCommands[]=
{
  "allspr", "normspr", "limspr", "ncol", 0 
} ;
int
MachineCommand(int argc, char *argv[], int *p)
{
  int J;
  
  if (KeyboardCommand(argc,argv,p))
    return 1;
  for(J=0 ; XtraCommands[J] ; J++)
    if(!strcmp(argv[(*p)],XtraCommands[J])) break;
  switch (J) 
    {
    case 0: NSprMax_M=NSprMax_C=32; break;
    case 1: NSprMax_M=4; NSprMax_C=8; break;
    case 2: (*p)++;
            if((*p)<argc) 
	      NSprMax_M=NSprMax_C=atoi(argv[(*p)]); 
	    else printf("%s: No number of sprites per line\n",argv[0]);
	    break;
    case 3: (*p)++;
	    if ((*p)<argc) { 
	      ncol=atoi(argv[(*p)]);
	      if ((ncol<2)||(ncol>256)) ncol=64;
	    }
	    else printf("%s: No number of colors supplied\n",argv[0]);
	    break;
    }
  return XtraCommands[J]!=NULL; 
}
  
/** InitMachine **********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int 
InitMachine(void)
{
  int K,L;
  int depth, screen_num;
  Window root;

  /* Reset all variables */
  memset(XKeyMap,0xFF,sizeof(XKeyMap));
  XBuf=NULL;JoyState=0x00;
  CurDisk[0]=CurDisk[1]=0;
  FastForward=0;

  /* Catch signals */  
  signal(SIGINT,OnInt);
  signal(SIGHUP,OnBreak);
  signal(SIGQUIT,OnBreak);
  signal(SIGTERM,OnBreak);

  isgcrand=isgcbw=isMyStipple=0; 
  wide=high=(UseZoom>1)?1:0; /* don't support Zoom>2 for now */
  Width=xWIDTH(wide);
  Height=xHEIGHT(high); /*start big*/

  if(Verbose & 1)
    printf("Initializing X11 drivers:\n  Opening display...");
  Dsp=XOpenDisplay(Dispname);

  if(!Dsp) { if(Verbose & 1) printf("FAILED\n");return(0); }
  if (do_sync) XSynchronize(Dsp, True);

  if (!ChooseVisual(&do_install, &do_mono)) 
      return 0;
  screen_num=VisInfo.screen;  depth=VisInfo.depth;
  root=RootWindow(Dsp, screen_num);
  if (!ChooseFormat(do_mono)) 
      return 0;
  InitColor(do_install, do_mono, &CMap); 

  {
    XSetWindowAttributes wat;
    
    if(Verbose & 1) printf("  Opening window...");
    
    wat.colormap=CMap;
    wat.event_mask=MyEventMask;
    wat.background_pixel=0;
/*    wat.background_pixmap=None; */
    wat.border_pixel=0;
/*    wat.border_pixmap=None; */

    Wnd=XCreateWindow(Dsp,root,0,0,Width,Height,0,depth,InputOutput,
		      VisInfo.visual, 
		      CWColormap|CWEventMask|CWBackPixel|CWBorderPixel, &wat);
  }
  SetWMHints();
  if(!Wnd) { if(Verbose & 1) printf("FAILED\n");return(0); }
  if(Verbose & 1) printf("OK\n");

  XMapRaised(Dsp,Wnd);
  if (! InitKeyboard(Dsp,Wnd)) return 0;

  gcrand=MakeGCrand(); isgcrand=1;
  if (do_mono)
    { gcbw=MakeGCbw(); isgcbw=1; }
  else
    gcbw=gcrand;

  XClearWindow(Dsp,Wnd);

  brdrstat[0].b_next=212; /* prevent infinite loop in PutImage() */
  brdrstat[0].b_width=0;
  brdrstat[0].b_C0=brdrstat[0].b_C1=0;
  for ( L=1 ; L < 212 ; ++L)
    brdrstat[L]=brdrstat[0];

  XAutoRepeatOff(Dsp);

#ifdef MITSHM
  if(UseSHM) /* check whether Dsp supports MITSHM */
    UseSHM=XQueryExtension(Dsp,"MIT-SHM",&ShmOpcode,&K,&L);
  if(UseSHM)
  { if(!(UseSHM=TrySHM(depth, screen_num))) /* Dsp might still be remote */ 
    {
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
      if(ximage) XDestroyImage(ximage);
      if(Verbose & 1) printf("FAILED..falling back\n");
    }
  } 
  if (!UseSHM)
#endif
  {
    long size;
    
    size=(long)sizeof(byte)*HEIGHT*WIDTH*bitpix/8;
    if(Verbose & 1) printf("  Allocating %ld %s RAM for image...",
			   (size&0x03ff) ? size : size>>10,
			   (size&0x03ff) ? "bytes" : "kB");
    XBuf=(byte *)malloc(size);
    if(!XBuf) { if(Verbose & 1) printf("FAILED\n");return(0); }

    if(Verbose & 1) printf("OK\n  Creating image...");
    ximage=XCreateImage(Dsp,VisInfo.visual, do_mono ? 1 : depth,
                        do_mono ? XYBitmap : ZPixmap, 
			0, XBuf, WIDTH, HEIGHT, 8, 0);
    if(!ximage) { if(Verbose & 1) printf("FAILED\n");return(0); }
  }
  if(Verbose & 1) printf("OK\n");
  if(SyncFreq) SetIFreq(SyncFreq);

#ifdef SOUND
  /* Initialize sound */   
  if(InitSound(UseSound,Verbose))
    SetChannels(SndVolume,SndSwitch);
#endif SOUND

  return(1);
}

/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** Keyboard matrix.                                        **/
/*************************************************************/
void 
Keyboard(void)
{
  XEvent E;
  int sleeping;


#ifdef SOUND
    /* We used to flush the sound stream here */
#endif
  sleeping=0;
  while (XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E))
  {
    /* allow to process one keystroke at the time */
    if ( OnKeyEvent((XKeyEvent*)(&E)) ) break; 
  }

  /* process other events */ 
  do {
    if(CFlag) { DOption(); CFlag=0; }
    while (XCheckWindowEvent(Dsp,Wnd,OtherMask,&E) && (ExitNow==0))
      {
	switch(E.type)
	  {
	  case FocusOut:
	    sleeping=SaveCPU;
	    if (Verbose & 2) puts("FocusOut");
	    OnFocusOut((XFocusChangeEvent *)(&E));
#ifdef SOUND
	    StopSound(); /* Silence the sound        */
#endif
	    break;
	  case FocusIn:
	    sleeping=0;
	    if (Verbose & 2) puts("FocusIn");
	    OnFocusIn((XFocusChangeEvent *)(&E));
#ifdef SOUND
	    ResumeSound();/* Switch the sound back on */
#endif
	    break;
	  case Expose:
	    if (! E.xexpose.count)
	      PutImage();
	    break;
	  case MapNotify:
	    Mapped=1;
	    if (Verbose & 2) printf("Mapped\n");
	    break;
	  case UnmapNotify:
	    Mapped=0;
	    if (Verbose & 2) printf("UnMapped\n");
	    break;
	  case ConfigureNotify:
	    if ((Width!=E.xconfigure.width)
		|| (Height!=E.xconfigure.height)) {
	      Width =E.xconfigure.width;
	      Height=E.xconfigure.height;
	      wide=(Width>=WIDTH+32);
	      high=(Height>=HEIGHT+32);
	      if (Verbose & 2) 
		printf("Window resized to %d by %d\n",
		       Width, Height);
	      PutImage();
	      setwidth(wide);
	    }
	    break;
	  default:
	    if (Verbose & 2) printf("E.type=%d\n",E.type);
	    break;
	  }
      }
    if (XCheckTypedWindowEvent(Dsp, Wnd, ClientMessage, &E)) {
      if((E.xclient.format==32)
	 && (E.xclient.message_type==WM_PROTOCOLS_atom) 
	 && (E.xclient.data.l[0]==DELETE_WINDOW_atom)) {
	ExitNow=1;
      }
    }
  } while (sleeping && (ExitNow==0));
}

